



pip check
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest -v tests --ignore=tests/test_unstructure_collections.py --ignore=tests/test_preconf.py --ignore=tests/test_typeddicts.py
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
